
;=======================================================================
;file-dialog and saver/loader routines should be in the lower memory area to avoid clash with ROM/RAM-on/off banking
;due to the lack of individual BASIC/KERNAL ROM/RAM switching, saver routines should turn ROM on/off for each bytes/blocks to transfer
;ROM routine jumptable addresses
KERNAL .block
.if (MACHINE<2)
 IRQ = $FCB3
 SCREENPOS = $D839 ;read/write current screen/cursor position ; input: Carry=0:set, Carry=1:get, input/output: Y-register: X-coordinate, X-register: Y coordinate
 CURSORLEFT= $DD1C ;step cursor left
 GETIN = $EBD9 ;get character from keyboard
 CHKRUNSTOP = $FFE1 ;returns 0 (Z-flag=1) if RUNSTOP is pressed
.fi
SETNAM = $FFBD ;Set file name parameters. Input: A = File name length; X/Y = Pointer to file name. (Real address: $FDF9)
SETLFS = $FFBA ;Set file parameters. Input: A = Logical number; X = Device number; Y = Secondary address. (Real address: $FE00)
OPEN   = $FFC0 ;Open file. (Must call SETLFS and SETNAM beforehands.) Used registers: A, X, Y. (Real address: ($031A), $F34A)
CLOSE  = $FFC3 ;Close file. Input: A = Logical number. Used registers: A, X, Y. (Real address: ($031C), $F291.)
CHKIN  = $FFC6 ;Define file as default input. (Must call OPEN beforehands.) Input: X = Logical number. Used registers: A, X. (Real address: ($031E), $F20E)
CHRIN  = $FFCF ;(BASIN)Read byte from default input (for keyboard, read a line from the screen). (If not keyboard, must call OPEN and CHKIN beforehands.) Output: A = Byte read. Used registers: A, Y. (Real address: ($0324), $F157.)
CHKOUT = $FFC9 ;Open a channel for output (Registers affected:  A, X), input: A=data, X = logical file number , Real address: ($0320), $F250.
CHROUT = $FFD2 ;(BSOUT) Write byte to default output. (If not screen, must call OPEN and CHKOUT beforehands.) Input: A = Byte to write. (Real address: ($0326), $F1CA)
CLRCHN = $FFCC ;Close default input/output files (for serial bus, send UNTALK and/or UNLISTEN); restore default input/output to keyboard/screen. Used registers: A, X. (Real address: ($0322), $F333)
SCINIT = $FF81 ;Initialize VIC; restore default input/output to keyboard/screen; clear screen; set PAL/NTSC switch and interrupt timer. Used registers: A, X, Y. (Real address: $FF5B.)
SAVE   = $FFD8 ;SAVE. Save file. (Must call SETLFS and SETNAM beforehands.) Input: A = Address of zero page register holding start address of memory area to save; 
               ;X/Y = End address of memory area plus 1. Output: Carry: 0 = No errors, 1 = Error; A = KERNAL error code (if Carry = 1). (Real address: $F5DD.)
LOAD   = $FFD5 ;LOAD. Load or verify file. (Must call SETLFS and SETNAM beforehands.) Input: A: 0 = Load, 1-255 = Verify; X/Y = Load address (if secondary address = 0).
               ;Output: Carry: 0 = No errors, 1 = Error; A = KERNAL error code (if Carry = 1); X/Y = Address of last byte loaded/verified (if Carry = 0). Used registers: A, X, Y. (Real address: $F49E.)  
READST = $FFB7 ;Fetch status of current device, value of ST variable. (For RS232, status is cleared.),No Input , Output: A = Device status., Used registers: A.,Real address: $FE07.
.bend

ddevice=8 ;default disk device

filedialog	.proc			;input:Accu - extension read directory, typing filename, return filename in 'namebuf'
maxnamelen=12 ;maximum filename-length to display
extlen = 4 ;extension-length (with dot)
namesep= $22 ; " filename-delimiter character code
RETURNcode=13 ;return/enter (carreturn) charactercode  (signs end of directory)
dirname .byte $24		;.asc "$" ;$ is the name of disk-directory


.if (FILEDIALOG_SUPPORT>0)

dirxpos= 1 ;filelist X screen position
dirypos= 4 ;filelist Y screen position
namecolums= 3 ;number of filename-columns (screenwidth divided by max. filename-length
dirvmem= VRAM+dirypos*40+dirxpos ;Video-RAM base memory address of filelist
disknam= VRAM+40*24+5 ;Video-RAM place of diskname $ disk-ID (disk-header)
.if (EMULATOR!=0)
dirnamp = VRAM+40*23+(40-12)
.fi
diskfree=VRAM+40*24+35 ;place of free disk-place display
filenamp=VRAM+10 ;filename-position on screen
authorip=VRAM+40*2 ;author-info screen-position
filelistheight=19 ;height of file-lister/selector subwindow
filelstmax= filelistheight*namecolums ;maximum length of filelist that can be displayed on 1 screen

fdialog .if (EMULATOR==0)
        jsr initfiler ;this entry point doesn't pull the directory in automatically, need to press F1 for that
        jsr clrfsel ;sets filecnt to 0, so F3 won't run anything if no F1 pressed before
        jmp typesel
        .else
        lda #0
        sta not1st+1
        .fi

reredir jsr initfiler

readdir lda #$01		;	filename-length is 1
		ldx #<dirname	;	lo byte pointer to file name. ("$")
		ldy #>dirname	;	hi byte pointer to file name. ("$")
		jsr inifnam		;init filename memory position and length
		jsr KERNAL.OPEN		; - call open
		bcs NOredir		;if error
		jsr KERNAL.READST
		bne NOredir		;if error
		ldx #$01		;	file #1 ;if no error (device is present), just go forward
		jsr KERNAL.CHKIN		; - call chkin to set input file #.
;read and display disk-header
		jsr KERNAL.CHRIN
		cmp #$C3		;normally first byte is $01, VICE gives $C3 if device is not present
		bne +			;if Carry remained set, no device is present
NOredir	jsr nodevic		;if error opening device, notify user
		jmp typesel		;and skip filelist-making, still give a chance to return to e.g. editor and change drive-number
+		ldx #6
-		jsr KERNAL.CHRIN		;read through some unneeded characters (amount given in X)
		dex
		bne -
		ldx #0			;read diskname & ID
		stx dirready	;signs if directory fully read
-		jsr KERNAL.CHRIN
		beq listfil
        cmp #$60
        bcc +
        and #$1F
+		ora #$80
        cpx #(EMULATOR==0)? 24:35         ;don't allow too long diskheader to be displayed (crop beginning to max. length) cpx #24			;don't allow too long diskheader to be displayed (crop beginning to max. length)
		bpl +
		sta disknam,x	;standard 'jsr CHROUT' is not good for special display modes
+		inx
		bne -
;parse and display directory for filesizes and names
listfil jsr clrfsel		;clear file-selector area, init file-listing variables
flistlo ldx #3			;file-lister loop
-		jsr KERNAL.CHRIN		;get filesize (last one will be filesize in blocks)
		sta sizebuf,x	;sizebuf might be displayed or filesizes stored in a filesize-list to help user calculate
		dex
		bpl -
-		jsr KERNAL.CHRIN		;check for beginning of filename (1st " delimiter)
		cmp #RETURNcode ;return/enter signal - check for end of dir/filelist
		beq closdir		;if end of filelist, close device
		cmp #namesep
		bne -			;search for left " filename-delimiter
		ldy #0			;start checking&copying filename
-		jsr KERNAL.CHRIN
		cmp #RETURNcode ;return/enter signal - check for end of dir/filelist
		beq closdir		;if end of filelist, close device
		cmp #namesep	;check for filename end (right " filename-delimiter)
		beq frowend		;find $00 row-delimiter
		sta namebuf,y	;loads actual filename to buffer
		iny
		cpy #17			;maximum allowed filename-length for C64/Plus4 devices
		bne -
frowend jsr KERNAL.CHRIN		;check for $00 - end of line of filename, then comes beginning of next line pointer and filesize/freeblocks info
		bne frowend
chekext lda filecnt
		cmp #filelstmax ;check displayable filelist max. length
		bcs typesel		;if reached all displayable content, go to type/select name or read next dir. page
		;or bcs flistlo ;if reached, no further displaying, just checking rows for filesizes, etc.
		dey				;check name for .SWM/.SWI extension
        .if (EMULATOR!=0)
        bmi flistlo ;empty foldername not to display
        lda namebuf+0
        cmp #TZEMU_FOLDER_CHAR
        beq dispnam ;folder?
        .fi
		ldx #extlen-1	;check 4 character long (with dot) filename-extension to filter on filetypes (.SWM/.SWI)
-		lda namebuf,y
		cmp tzftype,x
		bne flistlo		;if there was any difference between extension and end of filename-string, don't display filename
		dey
		bmi flistlo		;if filename is empty (only extension), no point to display
		dex
		bpl -
dispnam lda namebuf,y
		cmp #$20		;is last character SPACE? -then sign it with a delimiter character
		bne dispnlp		;if not SPACE, simply display filename
		lda #$20+$40	;the invisible space will sign real end of 'space-ended' filename to loader/saver routine
		bne +			;jump
dispnlp lda namebuf,y	;filename-displayer loop
+		sta (diszp),y
		dey
		bpl dispnlp
		inc filecnt		;count displayed filenames
nextrow jsr cycfcol		;cycle filer-column counter, Y output is tabbing amount ;go to next screenplace for next filename
		jmp flistlo		;back to read next line

closdir inc dirready	;sign that directory fully read
.if (EMULATOR==0)
		lda sizebuf+1	;display free kilobytes on disk in decimal format (divide blocks by 4)
		lsr sizebuf+0	;divide by 2
		ror
		lsr sizebuf+0	;divide by 2 (4 in sequence)
		ror
		ldx #100		;convert
		jsr divideX
		sty diskfree+0
		ldx #10
		jsr divideX
		sty diskfree+1
		tax
		lda hexdigi,x
		sta diskfree+2
.fi
		jsr clsfile

;-------type filename or author info / select file to load/save--------------
typesel ;sei				;type/select filename & description
        jsr ROMdisa
        jsr waitkeyrelease
		lda #keymode_filer	;set filename-typer mode in key-handler
		sta keyermode
defawin lda #0			;self-written - default filer-window (selector is preferred in case of loader)
.if (EMULATOR!=0)
not1st  ora #selfmod
.fi
		beq +
		jsr setsele		;set filename-selector subwindow as default for loader
		jmp fkeylop
+		jsr setfnam		;set filename-typer subwindow by default
fkeylop lda #$CC
		cmp vscanpos    ;constant 50Hz timing loop for keyboard-handler and cursor-displayer
		bne *-3
		lda #0
		sta fkeybout	;reset keyhandler output
		jsr keyer   	;handles key-repeating and name-mode
		jsr disfcur		;display filer-cursors
		lda fkeybout	;keyhandler-output
		cmp #keychr.fdir ;$f8	;F1 key check
		bne +			;if F1 pressed, re-read directory
f1press .if (EMULATOR!=0)
        sta not1st+1
        .fi
f1pres2	jsr regauth		;clsfile ;close dir., save typed author-info to re-read filelist
		jmp reredir		;readdir ;re-read directory/file-list
+		cmp #keychr.fpdn ;$f9		;F3/UP(PageDown) key check
		bne +			;if F3/Pagedown pressed, flip to next page of disk-directory (if exists)
		lda dirready	;check, if directory fully read
		bne lopover		;if directory read entirely, no further paging
        lda filecnt
        beq + ;no F1 pressed before or no filelist -> F3 shouldn't take effect
		jmp listfil		;further read filelist
+		cmp #keychr.fexi ;$fe		;if stop/esc pressed, returning to editor (abort file-operation)
		bne +			;$fe/$ff
		jsr regauth		;clsfile ;close directory-file - register author-text
        jsr ROMdisa
        lda #keymode_menu
        sta keyermode
		sec				;cause file-dialog abort in main routine (restart editor)
		rts
+		cmp #keychr.facc ;$ff		;if RETURN/Enter pressed, take action (load/save)
		bne lopover
		ldy filerwindow
.if (EMULATOR!=0)
        beq notdir
.fi
		cpy #1
		beq +
.if (EMULATOR!=0)
        lda dirnamp+0 ;namebuf+0
        and #$7F
        cmp #TZEMU_FOLDER_CHAR ;$7F ;folder listed by SW-emulator?
        bne notdir
        jsr clsfile
        ldx #<dirnamp
        ldy #>dirnamp
        lda #12
        jsr inifnam ;KERNAL.SETNAM ;inifnam ;iniwfn
        jsr $FF40 ;KERNAL.CHDIR ;artificial folder-change function for SWemu
        lda #$F8
        jmp f1press
notdir  ;not directory, so load/save file
.fi
		jsr regname		;register filename in 'namebuf' and author-info
        jsr ROMdisa
        lda #keymode_menu
        sta keyermode
		clc				;cause load/save in main routine, if return pressed window 1 or 2 (filename/file-selector)
		rts
+		jsr setfnam		;if window 1, go back to filename-typer
lopover jmp fkeylop

;-----------------------------------------------------------------------
initfiler
        sei				;init ROM $ VICbank and read disk directory to the screen
        lda #$1b        ;leave extended-colour mode
        sta vmodeyscroll
        lda #filerfgcol
        jsr clearscrn
        lda #filerbgcol
        sta bgcol
        lda #filerbordercol
        sta bordcol
        jsr ROMena
        lda #rasterIRQdisable
        sta irqctrl ;disable raster-IRQ
		ldy #rowlen-1	;draw UI base
-		lda filetx1,y
		sta VRAM+0,y;filename and type (extension)
		lda #$62+$80	;$78  ;$71 ;$68
		sta VRAM+rowlen,y;separator line 1
		sta VRAM+rowlen*3,y;separator line 2
		sta VRAM+rowlen*23,y;separator line 3
		lda authorinfo,y
skipaut sta authorip,y	;author info and title
		lda filetx2,y
        .if (EMULATOR!=0)
        cpy #size(filetx2)
        bcs +
        .fi
		sta VRAM+rowlen*lastrow,y;diskname and free space
+		dey
		bpl -
        jmp smallch

;----------------------------------file-dialog subroutines----------------------------------
filenamer				;handles filename-typing/selection (called from keyer/nametyper routine in keyhandler.inc, char passed in Accu, keycode in Y)
		sta fkeybout	;for main routine
		cmp #$fa		;Control/TAB keys - cycle filer sub-windows
		bne filkeys
		ldy filerwindow
		bne +			;if filename-subwindow
		iny				;set Y to 1
		lda editauth	;if loader, no author-info editing, skipping that
		beq setsele		;if player.authorinfo editing disabled (=0) go straight to selector
		lda #authorinfo_len-1
		sta filerlenzp	;length of input-field
		lda #<authorip
		ldx #>authorip
		jmp setyfld		;set typer field
+		cpy #1
		beq setsele		;if 1 set filename-selector subwindow
		clc
		jsr diselcu		;first clear selector-cursor
		jmp setfnam		;then cycle back to filename-typer window
setsele lda filecnt		;set file-selector
		beq setfnam		;if filelist is empty, don't go to selector, go back to / stay at filename-typer instead
		lda #<dirvmem
		ldx #>dirvmem
		jsr inicptr		;set video-pointers
		sec				;set display-mode for cursor
		jsr diselcu		;display file-selector cursor
		ldy #2			;set to 2
		bne setyfld		;jump
setfnam lda #maxnamelen-1
		sta filerlenzp	;length of input-field
		lda #<filenamp
		ldx #>filenamp
		ldy #0
setyfld sty filerwindow ;branched to here if filer-window was 2
setyfl2 jsr inidptr		;takes A , X
		lda #0
		sta filerposzp	;reset cursor-position
		sta filercolumnzp	;counts filename-selector columns
		rts

filkeys ldx filerwindow ;handle keys depending on subwindow---------------------------------
		cpx #2
		bne typefld		;if not file-selector window, go to typer ;if file-selector subwindow, use cursor keys and Return only
        cmp #keychr.fcdn ;$81
        beq fiselUD
		cmp #keychr.fcup ;$82		;cursor up/down
		bne +
fiselUD	jsr fiselLR		;3 columns
		jsr fiselLR
		jmp fiselLR
+		cmp #keychr.fcle ;$83		;cursor left/right
		beq fiselLR
        cmp #keychr.fcri ;$84
        bne endfiln
fiselLR clc				;clear cursor
		jsr diselcu		;clear file-selector-cursor
		jsr fiselec		;change cursor-pointer if possible
		sec				;draw cursor
		jmp diselcu		;draw file-selector-cursor
fiselec .if (MACHINE<2)
        lda fkeybout	;the effective algorithm itself to select file
        cmp #keychr.fcle ;$83
		beq fselbwd
        cmp #keychr.fcup ;$82
        beq fselbwd
        .else
        lda shiftsig
        bne fselbwd
        .fi
fselfwd ldy filerposzp
		iny
		cpy filecnt
		bcs endfiln
		inc filerposzp

cycfcol inc filercolumnzp	;column-tabbing counter
		ldx filercolumnzp
		cpx #namecolums
		bne +
		ldx #0
		stx filercolumnzp
+		ldy colutab,x
		jmp incompt

fselbwd lda filerposzp	;filename-count
		beq endfiln
		dec filerposzp
		ldx filercolumnzp
		ldy colutab,x
		dec filercolumnzp	;cycle filer-column counter
		bpl +			;if column-turnaround, compensate non-integer divisibility of screen width per filename-columns
		ldx #namecolums-1
		stx filercolumnzp
+		jmp decptrs

typefld ldy filerposzp	;handle typing fields------------------------------------------------
		cmp #keychr.fhom ;$85
		bne +			;Home-key?
		lda #0
		sta filerposzp
		rts
+		cmp #keychr.fins ;$86		;insert?
		beq +
		cmp #keychr.fdel ;$80		;backspace?
		bne cursidw
		lda shiftsig	;Insert? (Shift+Backspace)
		beq bkspclp
+		ldy filerlenzp
-		cpy filerposzp
		beq clrcurp
		dey
		lda (datzp),y   ;shift letters forward
		iny
		sta (datzp),y
		dey
		bpl -
clrcurp lda #$20
		sta (datzp),y
endfiln rts
bkspclp jsr fcurbwd
		ldy filerposzp
-		cpy filerlenzp
		beq clrcurp		;BACKSPACE character to last cursorplace
		iny
		lda (datzp),y   ;shift letters back
		dey
		sta (datzp),y
		iny
		jmp -
cursidw .if (MACHINE<2)
        cmp #keychr.fcri ;$84		;cursor-sideways keys?
		beq fcurfwd
		cmp #keychr.fcle ;$83
		bne norchar
        .else
        cmp #keychr.fcri
        bne norchar
		lda shiftsig
		beq fcurfwd
        .fi
fcurbwd cpy #0			;move cursor backwards in name-field
		beq endfiln
		dec filerposzp
		rts
norchar cmp #0
		bmi endfiln		;avoid typing characters for remaining control characters
		sta (datzp),y;normal character branch
fcurfwd cpy filerlenzp	;move cursor forward (if possible)
		bpl endfiln
		inc filerposzp
		rts

;--------------------------------------------------------------------------------
disfcur					;display cursor over typed data
		jsr clrcurs		;clear type-fields from cursor invertions
		lda filerwindow
		cmp #2			;selector?
		beq +
		ldy filerposzp	;display text-typer cursor
		lda (datzp),y
		ora #$80
		sta (datzp),y
+		rts

diselcu
.if (EMULATOR!=0)
         php
         ldy #0
         lda (diszp),y
         and #$7F
         eor #TZEMU_FOLDER_CHAR ;folder?
         sta dirstor+1
         plp
.fi
        ldy #maxnamelen-1;display/clear file-selector cursor (action depends on Carry-bit, displays if Carry=1, clears if Carry=0)
-		lda (diszp),y
		and #$7f		;clear cursor
		bcc +			;if Carry bit set
.if (EMULATOR!=0)
dirstor ldx #selfmod
        beq setcurs ;=jmp
.fi
		sta filenamp,y
setcurs ora #$80		;set cursor
.if (EMULATOR!=0)
        sta dirnamp,y
.fi
+		sta (diszp),y
		dey
		bpl -
		rts
clrcurs ldy #maxnamelen-1;clear type-fields from cursor invertions
-		lda filenamp,y
		and #$7f
		sta filenamp,y
		dey
		bpl -
		ldy #authorinfo_len-1;clear cursor
-		lda authorip,y
		and #$7f
		sta authorip,y
		dey
		bpl -
		rts

;--------------------------
regname jsr clrcurs		;start by clearing type-fields from cursor invertions
		ldy #maxnamelen-1;copy/register typed/selected filename and author-info (latter only if editauth is on
-		lda filenamp,y	;loop search back to first non-Space character
		cmp #$20		;check for spaces (if filename is 'spaced;, $60 Space will be noticed)
		bne +			;if not space coming, we found end of filename (without extension), and we can add the extension to it
		dey
		bpl -			;if not looped here, empty filename
+		iny
		sty filercolumnzp	;copy filename before last space
		beq +			;zero filename
		ldy #0
-		lda filenamp,y
		sta namebuf,y
		iny
		cpy filercolumnzp	;stored length of filename
		bne -
+		ldx #0			;add extension
-		lda tzftype,x	;read extension-string
		sta namebuf,y	;add it to namebuf.
		iny
		inx
		cpx #extlen
		bne -
		sty bufnlen+1	;register buffered name's length
regauth lda editauth
		beq clsfile		;if author-info being edited, register it in tune-header
		ldy #authorinfo_len-1;hardwired author-info length
-		lda authorip,y
		sta authorinfo,y
		dey
		bpl -
clsfile jsr ROMena
		lda #1			;	file # to close
		jsr KERNAL.CLOSE		; - so close it
		jmp KERNAL.CLRCHN		; - clear all channels

clrfsel	;clear file-selector area, init file-listing variables
        jsr ROMena
		lda #<dirvmem
		ldx #>dirvmem
		jsr inicptr		;set video-pointers
		ldx #0			;clear filelist-area
		stx filecnt		;counts displayed filenames
		stx filercolumnzp	;counts columns
fsclrlp ldy #rowlen-2
		lda #$20		;Space character
-		sta (datzp),y
		dey
		bpl -
		lda #rowlen		;one row forward
		jsr incptrs		;increase datzp	 ;iny -> Y=0 -> preserve diszp
		cpx #filelistheight
		bne fsclrlp
		rts

nodevic					;if OPEN failed, warn user and ask for retrial
		ldy #28			;display 'device not present'	(and later ask for overwrite-confirmation)
-		lda nodevicep,y
		sta VRAM+rowlen*lastrow,y
		dey
		bpl -
		rts ;jmp waitspace
		
filetx2 .text "DR:   .....................   FREE:...KB"
;filetx2 .text "DR:   F1,F3:DIR"
;.if (EMULATOR==0)
;        .text                ", CTRL:FIELD   FREE:...KB"
;.fi

;scratch .text "S0:" ;Scratch-signal (OPEN 15,8,15,"S0:file name":CLOSE 15) - will preceed filename when user want to overwrite a file (Save&Replace "@filename" is buggy in old 1541 drives)
namebuf .fill 16		;filename-buffer (for extension-checking and for disk-operations)
		.fill 10		;for safety reasons, if filename gets appended by commands

sizebuf .fill 4			;(byte 0 and 1) filesize-temp-storage, after reading dir., free blocks (last 2 bytes (row-pointers) not needed)
colutab .byte maxnamelen+2,maxnamelen+1,maxnamelen+1;column-tabbing values (to get screenwidth as sum) (14+13+13=40)
filecnt .byte 0			;amount of displayed files in directory
dirready .byte 0		;value is 1 if directory is fully read
editauth .byte 0		;0 - no author info display $ edition, 1 - autor info displayed & editable
filerwindow .byte 0		;window in file-menu: 0-filename-typer , 1-player.authorinfo-typer, 2-file-selector
fkeybout .byte 0		;keyboard-routine output of filer's keyboard handler - used by main loop to take some actions

;error messages in filer window, 29 character long each
nodevicep .text " NO DISK OR DRIVE PRESENT!!! "
fnotfound .text " FILE NOT FOUND! PRESS A KEY."
;filexists .text "ERROR! (EXISTS?) PRESS SPACE."
saveerror .text " SAVING ERROR 0! PRESS A KEY."

filenot ;jsr clsfile		;in case of error, close to allow further trials
		ldy #28			;Display 'file not found' message and wait for SPACE
-		lda fnotfound,y
		sta VRAM+lastrow*rowlen,y
		dey
		bpl -
		jmp waitkey

saverror ldy #28			;Display 'save error' message and wait for SPACE
-		lda saveerror,y
		sta VRAM+lastrow*rowlen,y
		dey
		bpl -
		jsr KERNAL.READST		;read status error-code
		clc
		adc #"0"		;shift numbering for status-number display to be PETSCII compatible
		sta VRAM+lastrow*rowlen+14;display status error-code
		jmp waitkey	;wait for key press/release

;!these pointer routines may be used by many other parts of the code , don't touch them, not even rearrange them!!!
inicptr					;lo-pointer in A, hi-pointer in X, set zeropage pointers - order of entry points is important!
		sta tempzp
		stx tempzp+1
ctempzp lda tempzp
		sta diszp
		ldx tempzp+1
		stx diszp+1
inidptr sta datzp
		stx datzp+1
		rts

incptrs					;increase source zeropage pointers by boundary size in A, and target pointer by Y value
		iny
incopt2 clc
		adc datzp
		sta datzp
		bcc incompt
incptnp inc datzp+1
incompt tya
incmpt2 clc
		adc diszp
		sta diszp	;set compressed data pointer to next position
		bcc +
		inc diszp+1
+		inx
		rts

decptrs tya
deccptr sta astore4+1
		lda diszp
		sec				;input: Accu -decrease 16bit player.tunedata-pointers by value given in a
astore4	sbc #selfmod
		sta diszp
		bcs +
dechptr dec diszp+1
+		dey
		rts

filetx1 .text "FILENAME:             "
		.byte $75		;the ruler between filename and extension display
		.text "(FILETYPE: *";. 'tzftype' should follow it immediately
tzftype .text ".TZM)"	;OR-file type - string to search extension in filenames to filter on tune/instrument


divideX ;.proc
        stx divzp	;divide A by divisor X, remainder is in A for further division, Y gets charcode to display
		ldx #0
-		sec
		cmp divzp	;compare divisor
		bcc +
		sbc divzp	;substract divisor
		inx
		bne -			;jump
+		
		ldy hexdigi,x
retdivX rts
        ;.pend


;-----------------------------------------------------------------------
.else ;no filedialog-support? simple typing when saving/loading using KERNAL routines, but directory should be requestable with $
;tzftype .text ".TZM"
namebufx=1
namebufy=2
namebuf=VRAM+namebufy*rowlen+namebufx

fdialog sei
        lda #filerbgcol
        sta bgcol
        sta bordcol
        jsr smallch ;capitch ;smallch
        jsr hidecursor
        lda #<KERNAL.IRQ  ;jsr $FF84, jsr $FF8A does similar
        ldx #>KERNAL.IRQ
        sta $FFFE
        stx $FFFF
        jsr ROMena
        jsr KERNAL.CLRCHN
        lda #$71                 ;lda #$84 ;jsr KERNAL.CHROUT ;flash off
        sta $0114 ;exomizer ruined yellow colour, so restore it 
        lda #$05 ;white colour
        jsr KERNAL.CHROUT
        lda #filerfgcol
        jsr clearscrn
        jsr waitkeyrelease
        ldy #rowlen-1
-       lda filenametxt,y
        sta VRAM,y
        dey
        bpl -
        lda #namesep
        sta VRAM+2*rowlen
resetsp ldy #namebufx
        ldx #namebufy ;lastrow
        clc
        jsr KERNAL.SCREENPOS
        cli
        jsr KERNAL.GETIN
        bne *-3 ;empty buffer
ftypelp lda #namesep ;$22 ;display " as cursor-character
        jsr KERNAL.CHROUT
        jsr KERNAL.CURSORLEFT
        jsr KERNAL.GETIN
        beq ftypelp
        bmi ftypelp
        cmp #namesep ;$22
        beq ftypelp ;don't allow typing "
        sec
        jsr KERNAL.SCREENPOS
        cmp #$03 ;RUNSTOP? 
        bne +
        rts ;Carry-flag=1 return-value is set by 'cmp'   
+       cmp #$0d
        beq fntyped
        cmp #$14 ;delete?
        bne +
        cpy #namebufx
        beq ftypelp ;resetsp
        jmp ++
+       cmp #$20
        bcc ftypelp
        cpy #namebufx+maxfilenamelen
        beq ftypelp
+       jsr KERNAL.CHROUT ;BSOUT
        jmp ftypelp
fntyped dey
        beq dispdir
        sty bufnlen+1
regname dey
-       lda namebuf,y ;namebufx ;nonzero? set Carry-flag return-value accordingly
        cmp dirname ;$?
        beq dispdir
        cmp #$20
        bcs +
        ora #$40
+       sta namebuf,y
        dey
        bpl -
        clc
        rts
dispdir jsr clearscrn
        ldx #0
        ldy #0
        clc
        jsr KERNAL.SCREENPOS
        lda #1	    	;	filename-length is 1
		ldx #<dirname	;	lo byte pointer to file name. ("$")
		ldy #>dirname	;	hi byte pointer to file name. ("$")
		jsr inifnam		;init filename memory position and length
		jsr KERNAL.OPEN		; - call open
		bcs filenot ;nodevic		;if device-not-present error
		ldx #$01		;	file #1 ;if no error (device is present), just go forward
		jsr KERNAL.CHKIN		; - call chkin to set input file #.
dirdilp jsr KERNAL.READST
        bne endrdir
        jsr KERNAL.CHRIN
        cmp #namesep
        bne dirdilp
-       jsr KERNAL.CHRIN
		cmp #namesep	;check for filename end (right " filename-delimiter)
		beq +     		;find $00 row-delimiter
		jsr KERNAL.CHROUT
		bne -
+       lda #RETURNcode
        jsr KERNAL.CHROUT
        sei
-       lda #0 ;$7F
		sta keymatrix
        sta keylatch
 		lda keylatch		;wait for keys to be released, especially RETURN
		and #$7F
        cmp #$7F
		bne -
        jsr KERNAL.CHKRUNSTOP ;RUN-STOP pressed?  ;also clears keyboard-buffer
        bne dirdilp
endrdir lda #1
        jsr KERNAL.CLOSE
        jsr KERNAL.CLRCHN
        jsr waitkey
        jmp fdialog

saverror
filenot sei 
        lda #2
        sta bordcol
        jmp waitkey ;Carry-flag is set to 1 inside
.enc screen
filenametxt .text "eNTER fILENAME ($/EMPTY:DIR,rUNsT:BACK):"
.enc none
.fi ;end of FILEDIALOG_SUPPORT


.pend ;end of file-dialog procedure
;=======================================================================



clearscrn ;input: Accumulator - foreground-color
        sta clrcolo+1
        ldy #0
-       lda #" "
        sta VRAM,y
        sta VRAM+$100,y
        sta VRAM+$200,y 
        sta VRAM+$300,y
clrcolo lda #selfmod
        sta CRAM,y
        sta CRAM+$100,y
        sta CRAM+$200,y
        sta CRAM+$300,y
        iny
        bne -
        rts

capitch .if (MACHINE<2)
        lda #$C4
        sta bitmapaddr
        lda #$D0 ;normal (CAPITAL) character set (bit2)
        sta chsetaddr1MHz
        .else
        lda #$15 ;normal character set
        sta vid_ch_pos
        .fi
        rts

smallch .if (MACHINE<2)
        lda #$C4
        sta bitmapaddr
        lda #$D0+4 ;small character set (bit2)
        sta chsetaddr1MHz
        .else
        lda #$16 ;small character set
        sta vid_ch_pos
        .fi
        rts

waitkeypress
-       lda #$80
        cmp vscanpos ;synchronizing to frames provides button-debouncing
        bne *-3
        lda #$00
        sta keymatrix
        sta keylatch
        lda keylatch
        cmp #$ff
        beq -
        rts

waitkey jsr waitkeyrelease
        jsr waitkeypress
waitkeyrelease
-		lda #$80 ;<singlerast
        cmp vscanpos ;synchronizing to frames provides button-debouncing
        bne *-3
        lda #$00
		sta keymatrix
        sta keylatch
		lda keylatch		;wait for keys to be released, especially RETURN
		cmp #$ff
		bne -
        rts ;Carry=1 on exit

.if (TUNEINIT_SUPPORT>0 || EXAMPLE_TUNE_SUPPORT>0)
YesOrNo					;wait for Y/N and set Carry accordingly (1 if Y)
		sei
YNloop  ;.if (EMULATOR==1)
        ; jsr waitfrm
        ;.fi
        lda #$e7
		sta keymatrix
-       ;.if (EMULATOR==1)
        ; jsr waitfrm
        ;.fi
        sta keylatch
    	ldy keylatch
		cpy #($ff-2)	;Y key (or I key)?
		bne +
        jsr waitkeyrelease
		sec
		rts
+		cpy #($ff-$80)	;N (or V) key?
		bne -     		;key-testing loop
        jsr waitkeyrelease
		clc
		rts
;waitfrm .if (EMULATOR==1)
;        ldy #$70
;        cpy $d012
;        bne *-3
;        dey
;        bpl *-1
;        rts
;        .fi

RyouSur					;Ask 'are you sure', wait for Y/N and set Carry accordingly (1 if Y)
RyouXpos = 12
RyouYpos = 10
RyouCol = $71
        lda #$12
        sta ECMcapcol
		ldy #16
-		lda #"*"		;DECOR
		sta VRAM+rowlen*(RyouYpos+0)+RyouXpos,y
		sta VRAM+rowlen*(RyouYpos+3)+RyouXpos,y
		lda #RyouCol
		sta CRAM+rowlen*(RyouYpos+0)+RyouXpos,y
		sta CRAM+rowlen*(RyouYpos+1)+RyouXpos,y
		sta CRAM+rowlen*(RyouYpos+2)+RyouXpos,y
		sta CRAM+rowlen*(RyouYpos+3)+RyouXpos,y
		lda RyouSurTx,y
		sta VRAM+rowlen*(RyouYpos+1)+RyouXpos,y
		lda RyouSurT2,y
		sta VRAM+rowlen*(RyouYpos+2)+RyouXpos,y
		dey
		bpl -
		jmp YesOrNo
dpywmk0 ldy #27
-       lda trksubpos,y
        dey
        lsr
        sta VRAM+$3B7,y
        bcc -
        rts
.enc screen
RyouSurTx .text "* ARE YOU SURE? *"
.enc screen ;none
RyouSurT2 .text "*  (Y/N) (I/N)  *"
.enc none
.fi


hexdisp pha
        lsr
        lsr
        lsr
        lsr
        tax
        lda hexdigi,x
        sta xstore1+1
        pla
        and #$0f
        tax
        lda hexdigi,x
xstore1 ldx #selfmod
        rts


.if (MENU_SUPPORT>0 || FILEDIALOG_SUPPORT>0)
decdisp ;.proc			;decimal display - input:accu, output: in Accu and Y-register
		ldy #$2f
		sec
-		iny
		sbc #10
		bcs -
		adc #$3a
		rts
		;.pend
.fi


ROMena  .if (MACHINE<2)
        sta ROMon ;enable KERNAL and BASIC ROMs
        .else
        lda #$36  ;enable KERNAL ROM
        sta banksel
        .fi
        rts

ROMdisa sei
        .if (MACHINE<2)
        sta ROMoff ;disable KERNAL and BASIC ROMs
        .else
        lda #$35
        sta banksel ;disable KERNAL and BASIC
        .fi
        rts

hidecursor
        .if (MACHINE<2)
        lda #$ff ;hide cursor
        sta curposlo
        sta curposhi
        .else
        lda #0 
        sta $d015
        .fi
        rts

iniwfn	;init workfile name pointer and length, bank, and device,file,channel
        jsr ROMena
        jsr KERNAL.CLRCHN
bufnlen lda #selfmod	;self-written code - length of selected/typed filename
		ldx #<filedialog.namebuf
		ldy #>filedialog.namebuf
inifnam jsr KERNAL.SETNAM		;init filename pointer and length, and device,file,channel
devicen ldx #ddevice	;self-written, device # 8 is default (modifiable by user)
.if (MENU_SUPPORT>0 || FILEDIALOG_SUPPORT>0)
		txa
		jsr decdisp
		sty VRAM+$3c3
		sta VRAM+$3c4
.fi
		lda #1			;	file # 1
		ldy #0			;	channel # 0 - (in case of LOAD: if 1, load to address automatically, if 0, address can be given in X/Y)
		jmp KERNAL.SETLFS		; - call setlfs


.if (PLAYABLE_EXPORT_SUPPORT>0 || RUNNABLE_EXPORT_SUPPORT>0)
appendPRGW ldx bufnlen+1	;length of selected/typed filename with extension
		ldy #0
-		lda PRGwriTxt,y
		beq +			;end of string?
		sta filedialog.namebuf,x
		inx
		iny
		bne -
+		stx bufnln2		;adjust filename-length to be expanded with ",P,W"
		rts
PRGwriTxt .null ",P,W"	;the appended modifiers after the filename for PRG-type outputs
bufnln2 .byte 0			;the length of the appended (,P,W) filename for OPEN operation

OpenExp	;init workfile name pointer and length, bank, and device,file,channel
		jsr ROMena      ;init workfile name pointer and length, bank, and device,file,channel
        jsr KERNAL.CLRCHN
		lda bufnln2		;appended filename-length
		ldx #<filedialog.namebuf	;point to appended (,S,W or ,P,W) filename
		ldy #>filedialog.namebuf
		jsr KERNAL.SETNAM		;init filename pointer and length, and device,file,channel
		lda #1			;	file # 1
		ldx devicen+1	;self-written, device # 8 is default (modifiable by user)
		ldy #2			;	channel # 2 ;this can be used for SEQ file formats
		jsr KERNAL.SETLFS		; - call setlfs
		jsr KERNAL.OPEN		;open the file for writing
		bcc OKtoSav		;check if file exists, if exists, go forward with saving
ExitErr	jsr filedialog.saverror	;display save-error and wait for keypress
		jsr filedialog.clsfile		;in case of error, close to allow further trials
		sec  ;sign error to caller
		rts
OKtoSav	ldx #$01		;	file #1
		jsr KERNAL.CHKOUT		; - call chkout to set output file #.
		clc ;sign 'no error' to caller
		rts

SaveDat	sta datzp+0 ;startaddress (by A and X), end-address in 'savendL+1', 'savendH+1'
		stx datzp+1
		ldy #0			;reset Y to always be 0 while saving (no need to index, zp-pointers increase instead)
saveloop				;save effective file-content byte-by-byte
		jsr KERNAL.READST
		bne ExitErr ;error?
+		lda (datzp),y
		jsr KERNAL.CHROUT		;write byte to disk
		inc datzp+0
		bne +
		inc datzp+1
+		lda datzp+0	;check if end of file
savendL	cmp #selfmod
        bne saveloop
		lda datzp+1
savendH	sbc #selfmod
		bne saveloop
dontclo	ldx #selfmod ;$00		; filenumber 0
        bne expdone
		jsr KERNAL.CHKOUT		; call CHKOUT (reset output device to prevent keyboard-typings being written to disk)
		jsr filedialog.clsfile		;close file #1
expdone clc ;sign OK to caller
		rts
.fi



;--------------------- Loader/Saver/Exporter routines ------------------
;must be in the lower memory area to allow switching on ROM
loadWorkfile
        sei
        ;jsr ROMena
        jsr iniwfn
        ldx #<tuneheader
        ldy #>tuneheader
        lda #0
        jsr KERNAL.LOAD
        bcc +
        jsr filedialog.filenot     ; if carry set, a load error has happened
        jsr ROMdisa
        jmp loadDialog ;start over
+       stx tuneendptr+0 ;end of data is important info for depacker to start in-place backwards-decompression
        sty tuneendptr+1
        jsr ROMdisa
.if (RUNNABLE_EXPORT_SUPPORT>0)
        jsr initPlayTimes
.fi
        jsr depackworkfile
.if (INFOBAR_SUPPORT>0)
        jsr nofirst
.fi
        jmp restEdi ;retuEdi

.if (RUNNABLE_EXPORT_SUPPORT>0)
initPlayTimes ldx #maxsubtuneamount*2-1
        lda #0
        ;sta tuneheader.playtime0+0
        ;sta tuneheader.playtime0+1
-       sta SubtuneDurationsBCDminSec,x ;don't bring previous tune's durations into new tune
        sta (EXEstarter.SubtuneDurations-BASIC_START)+EXEstarter,x ;seconds
        dex
        bpl -
        rts
.fi


saveWorkfile
        jsr packworkfile ;lda#packmode_workfile jsr packtune ; sets tuneendptr
        sei
        ;jsr ROMena
        jsr iniwfn ;includes ROMena
        lda #<tuneheader
        ldx #>tuneheader
        sta tunebaseptr+0
        ;sta tuneheader.baseaddress+ ;done in depttmp already
        stx tunebaseptr+1
        ;stx tuneheader.baseaddress+1
        lda #<(tunebaseptr+0)
        ldx tuneendptr+0
        ldy tuneendptr+1
        jsr KERNAL.SAVE
        bcs +           ;if saving error (e.g. file exists)
        jsr KERNAL.READST
        ;cmp #KERNAL.IOerror.NO_ERROR ; test Accumulator for status-info
        beq getback     ;if no saving error
+       jsr filedialog.saverror
        jsr ROMdisa
        jsr depackworkfile ;
        jmp saveDialog
getback jsr ROMdisa
        jsr depackworkfile ;depack tune - input needed: tuneendptr (end of packed tune)
        .if (EMULATOR==0)
        jmp retMenu
        .else
         jmp menuJmp
        .fi


.if (INSTRUMENT_LOADSAVE_SUPPORT>0)
loadIns sei
        ;jsr ROMena
        jsr iniwfn
        ldx #<instemp
        stx compzp+0
        ldy #>instemp
        sty compzp+1
        lda #0 ;load, not verify
        jsr KERNAL.LOAD
        bcs + ;bcc restEdi ;retuEdi
        txa ;X&Y points right after last loaded byte
        sec
        sbc compzp+0
        sta inssize
        jsr ROMdisa
        ldx currins
        lda insptlo,x
        sta datzp+0
        lda inspthi,x
        sta datzp+1
        jsr depack1inst
        jmp softRst
+       jsr filedialog.filenot     ; if carry set, a load error has happened
        jsr ROMdisa
        jmp loadInsDialog ;start over

saveIns sei
        jsr ROMena
        jsr iniwfn
        jsr ROMdisa
        lda #<instemp 
        ldx #>instemp
        sta compzp+0
        sta idatzp+0 ;using idatzp to store start-addres for saving
        stx compzp+1 
        stx idatzp+1 
        ldx currins
        stx tuneheader.ins_amount ;even save empty instrument (almost) correctly
        jsr pack1inst
        jsr ROMena
        ldx compzp+0 ;endaddress lo
        ldy compzp+1 ;endaddress hi
        lda #<(idatzp+0) ;startaddress
        jsr KERNAL.SAVE
        bcs +           ;if saving error (e.g. file exists)
        jsr KERNAL.READST
        ;cmp #KERNAL.IOerror.NO_ERROR ; test Accumulator for status-info
        .if (EMULATOR==0)
        beq retMenu     ;if no saving error
        .else
         bne +
         jsr ROMdisa
         jmp menuJmp
        .fi
+       jsr filedialog.saverror
        jsr ROMdisa
        jmp saveInsDialog
.fi


;lower area's ROM==on version of returners
retMenu jsr ROMdisa ;return to menu
        lda #exitmode_justlaunch
		sta exitsig
		rts

retuEdi jsr retMenu ;return to editor
retuEd2 lda #keymode_editor
        sta keyermode
        rts

restIns lda #1
        sta currins
softRst lda #exitmode_softrestart
        sta exitsig
        jmp retuEd2

restTun jsr ROMdisa
        lda #0
        sta cursubtune
        jsr (SUBTUNE_SUPPORT!=0)? setsubt : player.initzak ;INIT, subtune in Accu
        jmp softRst

restEdi lda #exitmode_restart
        sta exitsig
        jmp retuEd2


.if (PLAYABLE_EXPORT_SUPPORT>0)
SavePlayable
		jsr ROMdisa
        lda #packmode_release
        jsr packtune
		jsr relocate
        jsr appendPRGW	;append PRG file Writing string to filename
		jsr OpenExp
        bcc +
        jsr revert0
restExp jmp exportPlayableDialog ;start all over if error
+       jsr ROMena        
        lda #<PLAYERADDR
		jsr KERNAL.CHROUT		;address lo-byte
		lda relocaddr
		jsr KERNAL.CHROUT		;address hi-byte
        jsr SaveTun ;handles 'unrelocate' and 'depacktune' too
        bcs restExp
        .if (EMULATOR==0)
        jmp retMenu ;retuEdi ;returnM ;jmp goBackE  
        .else
         jmp menuJmp
        .fi

.fi

.if (PLAYABLE_EXPORT_SUPPORT>0 || RUNNABLE_EXPORT_SUPPORT>0)
SaveTun 
 .if (MACHINE<2)
        ldy #0       
        ldx #20
        clc
        .if (EMULATOR==0)
        jsr KERNAL.SCREENPOS ;don't allow possible upcoming LOAD/SAVE prompt to overwrite the insnametemp-data
        .fi
 .fi
        lda tuneendptr+0
        ldx tuneendptr+1
        sta savendL+1
        stx savendH+1
        lda #0
        sta dontclo+1
		lda #<PLAYERADDR ;Set Start Address for saver loop
		ldx #>PLAYERADDR
        jsr SaveDat
revert0 php ;store Carry
revertp jsr ROMdisa
		jsr unrelocate
        lda #packmode_release
        jsr depacktune
		plp ;restore Carry
        rts
.fi

.if (RUNNABLE_EXPORT_SUPPORT>0)
SaveRunnable
        jsr ROMdisa
        lda #packmode_release
        jsr packtune
        lda #>EXEstarter.EXE_PLAYERADDR
        jsr reloca2
; .if (TMF_EXPORT_SUPPORT)
        ldy #0
-       lda #' '
        sta (EXEstarter.TMFheader.Title-BASIC_START)+EXEstarter,y
        iny
        cpy #(EXEstarter.TMFheader.Tracker - EXEstarter.TMFheader.Title)
        bne -
        jsr setTEDtitle
        jsr setTEDplayback
; .fi
appendn jsr ROMena
        jsr appendPRGW	;append PRG file Writing string to filename
		jsr OpenExp
        bcc +
        jsr revert0
restREx jmp exportRunnableDialog ;start all over if error
+       lda #<BASIC_START
		jsr KERNAL.CHROUT		;address lo-byte
		lda #>BASIC_START
		jsr KERNAL.CHROUT		;address hi-byte
        lda #<EXEstarter_end
        ldx #>EXEstarter_end
        sta savendL+1
        stx savendH+1
        lda #<EXEstarter
        ldx #>EXEstarter
        stx dontclo+1
        jsr SaveDat
        ;bcs restREx
        ;jsr ROMena
        jsr SaveTun ;handles 'unrelocate' and 'depacktune' too
        bcs restREx
        .if (EMULATOR==0)
        jmp retMenu ;retuEdi
        .else
         jmp menuJmp
        .fi

; .if (TMF_EXPORT_SUPPORT)
setTEDtitle
 .if (EMULATOR!=0)
.cerror (*>$E000), *
        jsr ROMena
        ldx #0
-       lda $A009,x ;MIDIbuf1,x ;from fake KERNAL, in SIDmaker it's filled continuously from PC RTC-date
        beq +
        sta (EXEstarter.TMFheader.ReleaseInfo-BASIC_START)+EXEstarter,x
        inx
        cpx #$1F
        bcc -
+       ;lda #0
        ;sta (EXEstarter.TMFheader.ReleaseInfo-BASIC_START)+EXEstarter,x ;terminating zero
        jsr ROMdisa
 .fi
        ldx #0          ;will show the position of ':' - if no separator, value will stay 0
        ldy #39
-       lda authorinfo,y
        cmp #":"        ;if there's a ':' separator, the string before will be handled as author-name
        bne +
        tya             ;only the 1st ':' will be used as split-separator
        tax
+       dey
        bpl -
        cpx #0
        bne splitTX     ;if ':' was found
noSplit ldy #32-1       ;SID-header text field max. size
-       ;lda #0          ;reset author-info text
        ;sta (EXEstarter.TMFheader.Author-BASIC_START)+EXEstarter,y
        lda authorinfo,y
        sta (EXEstarter.TMFheader.Title-BASIC_START)+EXEstarter,y
        dey
        bpl -
        rts
splitTX cpx #32-1
        bpl noSplit     ;if the separator is at illegal place (authorname would be longer than allowed 32
        ;lda #0
        ;sta (EXEstarter.TMFheader.Author-BASIC_START)+EXEstarter,x ;put '0' string-termination
        txa
        tay
-       lda authorinfo-1,y ;right before ':'
        sta (EXEstarter.TMFheader.Author-BASIC_START)+EXEstarter-1,y
        dey
        bne -
        ;ldy #0
-       lda authorinfo+1,x ;right after ':'
        cpy #0
        bne +
        cmp #' '
        beq skipspc ;skip the space separator after :
+       sta (EXEstarter.TMFheader.Title-BASIC_START)+EXEstarter,y
        iny
        cpy #32 ;-1       ;max text field-length in SID-header
        beq +
skipspc inx
        cpx #40-1
        bne -
+       ;lda #0          ;put '0' string-termination
        ;sta (EXEstarter.TMFheader.Title-BASIC_START)+EXEstarter,y
        rts

BCDminSecToBinary ;inputs: X:minutesBCD, A:secondsBCD,  outputs: X:secondsH, A:secondsL
        stx xstorer+1
        jsr BCDtoBinary
        pha
        lda #0
        sta binsecH+1
xstorer lda #selfmod
        jsr BCDtoBinary
        pla
        jmp +
-       clc
        adc #60 ;add 60 seconds for a minute
        bcc +
        inc binsecH+1
+       dex
        bpl -
binsecH ldx #selfmod
        rts

BCDtoBinary ;in: A, out: A,X
        sed
        ldx #-1
-       inx
        sec
        sbc #1
        bcs -
        cld
        txa
        rts

setTEDplayback
        lda tuneheader.subtune_amount
        sta (EXEstarter.TMFheader.NumberOfSongs-BASIC_START)+EXEstarter
        asl
        tay ;ldy #maxsubtuneamount
-       ldx SubtuneDurationsBCDminSec-2+0,y ;minutes
        lda SubtuneDurationsBCDminSec-2+1,y ;seconds
        cpy #2
        bne +
        ldx tuneheader.playtime0+0 ;minutes
        lda tuneheader.playtime0+1 ;seconds
+       jsr BCDminSecToBinary
        sta (EXEstarter.SubtuneDurations-BASIC_START)+EXEstarter-2+0,y ;seconds
        txa
        sta (EXEstarter.SubtuneDurations-BASIC_START)+EXEstarter-2+1,y ;minutes
        dey
        dey
        bne -
        ldy tuneheader.framespd
        lda frmspds,y   ;check framespeed-program lookup table in datatables.inc
        sta fspdsto+1
        dey
        sty (EXEstarter.exefspd+1-BASIC_START)+EXEstarter
        sty (EXEstarter.TEDfspd-BASIC_START)+EXEstarter+1
        beq + ;appendn     ;if 1x single-speed, don't disturb table
-       tya
        clc
fspdsto adc #selfmod    ;self-written - add base-position of current rasterspeed
+       tax
        lda EXEframerows,x ;get current frameraster from table
        sta (EXEstarter.EXEframes-BASIC_START)+EXEstarter,y ;graving into the executable-code
        dey
        bpl -
        lda #<EXEstarter.EXE_INITADDR
        ldx #>EXEstarter.EXE_INITADDR
        ldy tuneheader.framespd
        dey ;cpy #1          ;if framespeed is 1 don't use CIA starter code
        beq +
        lda #<EXEstarter.TIMERINIT
        ldx #>EXEstarter.TIMERINIT
+       sta (EXEstarter.TMFheader.InitAddress-BASIC_START)+EXEstarter+0
        sta (EXEstarter.TMFheader.DataAddress-BASIC_START)+EXEstarter+0
        stx (EXEstarter.TMFheader.InitAddress-BASIC_START)+EXEstarter+1
        stx (EXEstarter.TMFheader.DataAddress-BASIC_START)+EXEstarter+1
        sec
        sbc #<TMFHEADER_ADDRESS
        sta (EXEstarter.TMFheader.DataOffset-BASIC_START)+EXEstarter+0
        txa
        sbc #>TMFHEADER_ADDRESS
        sta (EXEstarter.TMFheader.DataOffset-BASIC_START)+EXEstarter+1
        lda #<EXEstarter.EXE_PLAYADDR
        ldx #>EXEstarter.EXE_PLAYADDR
        ldy tuneheader.framespd
        dey ;cpy #1          ;if framespeed is 1 don't use CIA starter code
        beq +
        lda #<EXEstarter.TIMERPLAY
        ldx #>EXEstarter.TIMERPLAY
+       sta (EXEstarter.TMFheader.PlayAddress-BASIC_START)+EXEstarter+0
        stx (EXEstarter.TMFheader.PlayAddress-BASIC_START)+EXEstarter+1
        lda #0               ;for VIDEO-frame timing of all subtunes
        ldy tuneheader.framespd ;framespeed of loaded tune (in header)
        dey ;cpy #1          ;if framespeed is 1 don't use CIA starter code
        beq +                ;single-speed?
        lda #2               ;for TED-timer based timing of all subtunes
+       sta (EXEstarter.TMFheader.TimingSource-BASIC_START)+EXEstarter
        lda TimerFrameLo,y
        ldx TimerFrameHi,y
        sta (EXEstarter.TMFheader.Timer1value-BASIC_START)+EXEstarter+0
        stx (EXEstarter.TMFheader.Timer1value-BASIC_START)+EXEstarter+1
retSTPl rts
;CIA-timing values for multispeed. ;Ian Coog told this formula: ((screen lines) * (cycles per line) / speed) - 1 
;17734 2x:8867, 4x:4433
TimerFrameHi .byte $45,$22,$17,$11, $0d,$0b,$09,$08 ;replaced by exporter to PAL or NTSC values
TimerFrameLo .byte $46,$A3,$17,$51, $da,$8b,$e5,$A8 ;replaced by exporter to PAL or NTSC values

; .fi
.fi


;----------------------- Lower area data/variables ---------------------
relocaddr .byte >PLAYERADDR ;relocation target address HI-byte
zeropaddr .byte playzp_default

.enc screen
hexdigi .text "0123456789ABCDEF"
.enc none
TZWext .text "TZW"
TZIext .text "TZI"
PRGext .text "TZP"
RUNext .text "TED" ;.text "TZR"
;TEDext .text "TED"

authorinfo_len=rowlen
authorinfo .fill authorinfo_len," " ;will be copied to player-variable area for playable, or end of packed data for workfile
authorinfo_end


;================== Player-program for runnable export =================


.if (RUNNABLE_EXPORT_SUPPORT>0)

TMFHEADER_ADDRESS = BASIC_START-2 ;for TED format
EXEstarter .block
EXE_INITADDR = EXE_PLAYERADDR+0
EXE_PLAYADDR = EXE_PLAYERADDR+3
EXE_MULPLAYADDR = EXE_PLAYERADDR+6
.cerror (start2>9999), "starter code address aka SYS-parameter can't be above 9999! Would misalign TMF-header."

TMFheaderFormat .macro               ;Offset:      default/info: ;TMF-format header within Hermit's custom TED-format container
 ;LoadAddress       .word BASIC_START   ;0,1        PRG load-address, always $1001 to be runnable from BASIC prompt (added automatically by PRG-saving)
 BASICnextLine     .word BASICend       ;2,3 next BASIC line's address ($100B)
 BASIClineNumber   .word RealTMFeader-TMFHEADER_ADDRESS ;4,5     BASIC SYS-command's line-number, OFFSET TO THE TED(TMF)-HEADER ($2020)
 BASICsysCommand   .byte $9E            ;6          BASIC SYS-command token ($9E)
 BASICsysParameter .text format("%d",start2) ;7..10  BASIC SYS-parameter, the address of runnable machine code in ASCII (4352=$1100)
 BASIClineEnd      .byte $00            ;11     $0B BASIC line-end ($00)
 BASICend          .word $0000          ;12,13  $0C BASIC program end ($00,$00)
 Filler            .fill 3, $00         ;14..16 $0E filler empty places (0,0,0)
 RealTMFeader ;TMF-only part comes
 Signature         .text "TEDMUSIC"     ;17..24 $11 "TEDMUSIC" MagicString
 SpecVersion       .byte 0              ;25     $19 Preliminary specification-version (0)
 DataOffset        .word EXE_PLAYERADDR-TMFHEADER_ADDRESS ;26,27  $1A Where the music (maybe with loadaddress) resides - relative to beginning of file
 DataAddress       .word EXE_PLAYERADDR ;28,29  $1C Where the music should be loaded/copied
 InitAddress       .word EXE_INITADDR   ;30,31  $1E Init-routine to start/set song, subtune in Accu. Init mustn't change IRQ
 PlayAddress       .word EXE_PLAYADDR   ;32,33  $20 Play-routine's address, don't disturb IRQ (0=RTED?)
 NumberOfSongs     .byte 1              ;34     $22 amount of subtunes ('1' if only 1)
 TimingSource      .byte 0              ;35     $23 0:PAL Vblank(50Hz), 1:NTSC Vblank(60Hz), 2:PAL Timer1, 3:NTSC Timer1, 4:PAL Timer1(set by init), 5:NTSC Timer1(set by init)
 Timer1value       .word 17734          ;36,37  $24 TED timer1 value (PAL: (1x:17734), 2x:8867, 4x:4433) (NTSC: (1x:14915), 2x:7457, 4x:3728)
 SongFlags         .byte %00000000      ;38     $26 bit4:FM?, bit3:AY?, bit2:DigiBlaster?, bit1:SIDcard?, bit0:Screen-off?
 FileFlags         .byte %00000001      ;39     $27 bit0:SongLength (duration) block is present (2 byte-words for all subtunes: seconds)
 DurationOffset    .word SubtuneDurations-TMFHEADER_ADDRESS ;40,41  $28 file-offset of the duration(songlength)-block
 Reserved          .fill 23, $00        ;42..64 $2A reserved for future use (padding to have coming textfields aligned to $10 in memory)
 Title             .fill 32, ' '        ;65..96 $41 Title of music
 Author            .fill 32, ' '        ;97...  $61 Author/composer of music
 ReleaseInfo       .fill 32, ' '        ;129..  $81 Release-info, recommended: [Year] [Group]
 Tracker           .text "TEDzakker-"   ;161..  $A1 Tracker/composer-tool that the music was created with
 TrackerVersion    .text Version,"-" ; "   "
 TrackerType       .text (PLAYERTYPE==0)?"normal": (PLAYERTYPE==1)?"medium": (PLAYERTYPE==2)?"light ": (PLAYERTYPE==3)?"extra ": (PLAYERTYPE==4)?"bare  ": (PLAYERTYPE==5)?"demo  " : ""
                   .text "            "
 ReservedInfoField .fill 64, ' '        ;193..  $C1 Reserved textfield for possible future use
 ;(257..=$101.. built-in starter/initializer routine for pure TMF,) duration-block, etc.
.endm

 .logical BASIC_START
TMFheader #TMFheaderFormat ;.word ss2,TMFHEADER_ADDRESS ;Year     ;.null $9e,^start2 ;SYS command      ;ss2 .word 0
start2  sei
        lda #exebgcol
        sta bgcol
        .if (MACHINE<2)
        lda #$08
        sta VRAMaddr
        lda #0
        sta VRAMposhi
        lda #$F8
        sta VRAMposlo
        lda #$C4
        sta bitmapaddr
        lda #$D0+4 ;small character set (bit2)
        sta chsetaddr1MHz
        .else
        lda #$03
        sta VICbank
        lda #$16 ;small character set
        sta vid_ch_pos
        .fi
        lda #$1b
        sta vmodeyscroll
        lda #$08 ;$48 for NTSC?
        sta cmodexscroll
        ldy #0
-       lda #exefgcol
        sta CRAM+$000,y
        sta CRAM+$100,y
        sta CRAM+$200,y
        sta CRAM+$300,y
        lda #" "
        sta VRAM+$000,y
        sta VRAM+$100,y
        sta VRAM+$200,y
        sta VRAM+$300,y
        iny
        bne -
        ldy #rowlen-1
-       lda expinfo,y
        jsr convPETSCII
        sta VRAM+lastrow*rowlen,y
        dey
        bpl -
        jsr dispAuthorInfo
playSub lda #0 ;default subtune ;later should be changeable with +/-
        jsr EXE_INITADDR ;EXE_PLAYERADDR+0
framcnt ldy #1 ;selfmod
        lda EXEframes,y
        cmp vscanpos
        bne *-3
        ldx #(MACHINE<2)? 16:8 ;beautify flicker
        dex
        bne *-1
        dey
        bpl +
exefspd ldy #1-1 ;selfmod ;written by exporter: framespd-1
+       sty framcnt+1
        bne +
        jsr playerCall
chSpace lda #$7F
        sta keymatrix
        sta keylatch
        lda keylatch
        cmp #$EF
        beq playSub
chEsc   lda #$BF
        sta keymatrix
        sta keylatch
        lda keylatch
        cmp #$EF
        bne framcnt
        jsr playerCall
        jsr playerCall
        jsr playerCall
        jmp framcnt
+       lda #exebarcol2
        sta bordcol
        jsr EXE_MULPLAYADDR ;EXE_PLAYERADDR+6
        jsr bordBck
        jmp framcnt

playerCall
        lda #exebarcol
        sta bordcol
        jsr EXE_PLAYADDR ;EXE_PLAYERADDR+3
bordBck lda #exebordcol
        sta bordcol
        rts

dispAuthorInfo
		ldy #rowlen-1
-		lda #" "
		sta VRAM+(rowamount/3)*rowlen,y
        lda keyinfo,y
        jsr convPETSCII
        sta VRAM+(rowamount-5)*rowlen,y
		dey
		bpl -
		ldy #rowlen-1
-		lda EXE_PLAYERADDR+(explayer.authorinfo-explayer.player_top),y
		cmp #" "
		bne + ;find character before last space
		dey
		bne -
+		sty columnzp
		lda #rowlen-1
		sec
		sbc columnzp
		tay
		lsr ;halve the distance -> center text
		tay
		ldx #0
-		lda EXE_PLAYERADDR+(explayer.authorinfo-explayer.player_top),x
		sta VRAM+(rowamount/3)*rowlen,y
		inx
		iny
		cpy #rowlen
		bne -
		rts
convPETSCII
        cmp #$60
        bcc +
        and #$1F
+       rts

;.enc screen
keyinfo .text "Press SPACE to restart,   ", MACHINE<2?"Esc":" <-", ": fast-fwd."
expinfo .text "It was created with TEDzakker-"
        .text (PLAYERTYPE==0)?"normal": (PLAYERTYPE==1)?"medium": (PLAYERTYPE==2)?"light": (PLAYERTYPE==3)?"extra": (PLAYERTYPE==4)?"bare": (PLAYERTYPE==5)?"demo" : ""
        .text "-",Version,"  "
;.enc none

;framespeed-program - overwritten/generated by exporter for specific rasterspeeds
EXEframes .byte <(exerast1-(PAL_SCANLINES/8)*range(0,7)) ;TABLE LOADED BY OTHER PARTS OF THE CODE (counts backwards)

SubtuneDurations .fill maxsubtuneamount*2, $00
.cwarn "EXEstarter-end=",*

*=EXE_PLAYERADDR-$20
TIMERINIT ldy #0
          sty frCount
          jmp EXE_INITADDR
TIMERPLAY dec frCount   ;SID CIA playing routine called by IRQ in player
          bmi TEDfspd
          jmp EXE_MULPLAYADDR
TEDfspd   lda #0 ;selfmod ;framespeed, written by exporter (0=1x, 1=2x, ...)
          sta frCount
          jmp EXE_PLAYADDR
frCount .byte 0
.cerror *>EXE_PLAYERADDR,"YOU MUST ENSURE MORE PLACE FOR TIMER-BASED TED PLAY-CODE!"

  .align $100 ;important, as player is only relocatable to $100 boundaries
EXE_PLAYERADDR ;EXE_PLAYERADDR=(BASIC_START/$100+1)*$100
 .here
.bend
EXEstarter_end
.fi ;end of RUNNABLE_EXPORT_SUPPORT

