
;============================== TEDZAKKER ==============================
;A music creation tool for the TED chip of the Commodore Plus/4 computer
;2023 Hermit Software Hungary (Mihaly Horvath)
;=======================================================================


.include "include/constants.asm"


;========================== BASIC STARTER-CODE =========================
BASIC_START=(MACHINE<2)? $1001: $0801

        *=BASIC_START
        .word ss,Year
        .null $9e,^start ;SYS command
ss        .word 0

;============================== START ==================================
start   sei
        ;lda devnum_zp ;store used device and put it between 8..15
        ;ora #$08
        lda #ddevice ;and #$0F ;exomizer destroys the drive-number information unfortunately
        sta devicen+1
        sta devnum_zp
        lda #PLAYERTYPE
        sta tuneheader.playertype ;this must be done by code, byte literal didn't work
        jsr ROMdisa
        .if (MACHINE==2)
        lda #3
        sta VICbank
        .fi
        jsr setirq
.if (MENU_SUPPORT>0)
        lda #keymode_menu ;keymode_editor ;starting with menu for the 1st time make sense
        sta keyermode  ;sta menusig+1 ;start with editor, not menu
        lda #0
        sta menupoint
        sta cursubtune
        jsr player.initzak ;INIT, subtune in Accu
        jmp menu
.fi
restart ;sei
        jsr ROMdisa
        lda #0
        sta cursubtune
        sta playmark+0
        sta playmark+1
        jsr setsubt ;jsr player.initzak ;INIT, subtune in Accu
        lda #1
        sta currins
        lda #$FF
        sta mutesolo+0
        sta mutesolo+1
restrt1 ;sei
        jsr ROMdisa
        lda #stopmode
        sta playmode
        lda #1
        sta jammode
        lda #0
        sta player.currnote+0
        sta player.currnote+1
        jsr player.softini ;jsr player.silence
launche ;sei
        jsr ROMdisa
        jsr iniscrn
        jsr waitkeyrelease
        lda #keymode_editor
        sta keyermode
        lda #exitmode_noexit
        sta exitsig
        jsr setirq
        cli


;============================ MAIN LOOP ================================
mainloop
menusig lda keyermode ;lda #selfmod
        cmp #keymode_menu
        bne refresh
 .if (HERMIDI_SUPPORT>0)
        sei
        lda HerMIDIswitch
        beq gomenu
        ;lda HerMIDI.StatusByte
        ;bne gomenu
        jsr HerMIDI.PowrOFF
        lda #0
        sta MIDIsig+1
 .fi 
gomenu  jmp menu
refresh jsr display
chexits lda exitsig
        ;cmp #exitmode_noexit
        beq mainloop
        cmp #exitmode_justlaunch ;e.g. exiting from help / menu
        beq launche
        cmp #exitmode_softrestart ;e.g. loading new instrument or partial musicdata-deletion (ins/tune) -> stop playback
        beq restrt1
        cmp #exitmode_restart ;e.g. loading new tune
        beq restart
        jmp restart


.if (MACHINE==2)
.align $40
cursprite
    .byte %01111111,%10000000,%00000000
   .for i=0,i<7,i=i+1
    .byte %10000000,%01000000,%00000000
   .next
    .byte %01111111,%10000000,%00000000
   .fill (21-9)*3+1
.fi


.include "include/fileoperations.asm"


;**************************** EXPORT-PLAYER ****************************
;the player that gets exported together with the packed TUNEDATA
;.cerror (*>=PLAYERADDR),"LOWER PART OF PROGRAM OVERLAPS PLAYER AND TUNEDATA ",*
;*=PLAYERADDR
.align $100
PLAYERADDR
.if (PLAYABLE_EXPORT_SUPPORT>0 || RUNNABLE_EXPORT_SUPPORT>0)
		.union	;unnamed union - creates an instance for both but at the same memory location in prg output, so 'player' will overwrite 'player2'
 .if (RELOCATION_SUPPORT>0)
		 .struct
explayer2 #playerm PLAYERADDR+$100,0,0 ;for comparison & detection of code to relocate
		 .ends
		 .struct
explayer3 #playerm PLAYERADDR,0,1 ;for comparison & detection of zeropage-accesses to modify
		 .ends
 .fi
		 .struct
explayer  #playerm PLAYERADDR,0,0
		 .ends
		.endu
;.else
;explayer=player
.fi


.include "include/tunedata.asm"


;=========================== main variables ============================
ptnwin=0 ;keep it as 0, so no need to have "cmp #ptnwin" in many places
oliwin=1
inswin=2
inamewin=3

cursubtune .byte 0
currptn    .byte 1,2 ;.fill CHANNELAMOUNT,1
currins    .byte 1
currwin    .byte ptnwin ;0:pattern-editor(keep it at 0!), 1:orderlist, 2:inst.table, 3:inst.name
curwinposx .byte 0,0,0,0
curwinposy .byte 0,0,0,0
octave     .byte 4
mutesolo   .fill CHANNELAMOUNT,$FF
stopmutesig .fill CHANNELAMOUNT,$FF

seqindex  .byte 0
ptnindex .fill CHANNELAMOUNT,0
insindex  .byte 0
playtime .word 0
minutes = playtime+0 ; .byte 0
seconds = playtime+1 ; .byte 0
.if (RUNNABLE_EXPORT_SUPPORT>0)
FullyPlayed .byte 1 ;zeroed at F1-initiated playback, set when a loop/end happens after playing the whole subtune, rising edge registers playtime
.fi

jammode  .byte 1
fulljam  .byte 0 ;jam everywhere, when jam-mode is ON (avoids accidental hex.value typing in ptn/instr.table)
notemode .byte 0
advance  .byte 1

keymode_editor=0
keymode_insname=1
keymode_filer=2
keymode_menu=3
keyermode .byte 0 ;0:normal(editor), 1:instr.name-typer, 2:file-dialog name-typer, 3:menu-mode
;helpmode .byte 0

playmode .byte stopmode
followmode .byte 0

.if (HERMIDI_SUPPORT>0)
HerMIDIswitch .byte 0
.fi

playmark .fill CHANNELAMOUNT,0 ;for F2-initiated play-from-mark entries
markcnt .byte 0
ffwdcnt .byte 0
fspdcnt .byte 0

;help from player regarding positions to display:
currtempo .fill CHANNELAMOUNT,0 ;written by player
playerpatt .fill CHANNELAMOUNT,0 ;written by player
playerpcnt .fill CHANNELAMOUNT,0 ;written by player
wftbpos   .fill CHANNELAMOUNT,0 ;written by player
freqtbpos .fill CHANNELAMOUNT,0 ;written by player
voltbpos   .fill CHANNELAMOUNT,0 ;written by player



;============================= INTERRUPT ===============================
setirq  sei
        lda #(MACHINE<2)? rasterIRQenable+(>singlerast) : $01
        sta irqctrl ;enable raster-IRQ and set raster-comparator MSB to 0
        lda #>singlerast
        sta rscn8st+1
        .if (MACHINE==2)
        lda #$1B+(>singlerast)*80
        sta vmodeyscroll
        lda #$7F
        sta $DC0D
        sta $DD0D
        lda $DC0D
        lda $DD0D
        .fi
        lda #<singlerast
        sta rastercmp ;set raster-IRQ line
        sta rscntst+1
        lda #<irq
        ldx #>irq
        sta $fffe        
        stx $ffff
        lda #<dummynmi
        ldx #>dummynmi
        sta $fffa
        stx $fffb
        lda #$FF
        sta irqstatus ;acknowledge and reset any IRQ (statusbits) that has already happened
        lda #0 
        sta fspdcnt
        rts


.include "include/irq.asm"


.include "include/keyhandler.asm"
;.cwarn(*>=$A000),"keyhandler end: ",*

.include "include/datawriter.asm"
;.cwarn(*>=$A000),"datawriter end: ",*


;======================== DISPLAY ROUTINES =============================
.if (MACHINE==2)
 .cerror(*>=$D000),"CODE-AREA BELOW I/O OVERLAPS C64 I/O AREA. MOVE SOME CODE ABOVE $E000, ",*
 .if ( * >= $D000-($FFFF-$E000) )
  *=$E000
 .fi
.fi


.include "include/displayer.asm"


.include "include/menu.asm"
;.cwarn(*>=$A000),"menu end: ",*


.include "include/tuneexport.asm"
.include "include/tuneimport.asm"


.include "include/midi.asm"


.if (MACHINE==2) 
 #pointertables
end_of_code
 *=$d000 ;data under I/O-area follows for C64
.fi


;============================== HELPTEXT ===============================
.if (HELPTEXT_SUPPORT>0)
;.enc screen
helptxt .block
.text "KEYS: F1/F2/F3: Play: begin/mark/pattern"
.text "----F4/RUNSTOP: Stop/Continue(C=:follow)"
.text "Sh+P/O/I F5-F8: Patt/Orderlist/Inst/Menu"
.text " Shift + SPACE: Put mark / Play from pos"
.text "ESC, C=/Sh+ESC: FastForward,(Auto)Follow"
.text "CONTROL, SPACE: Tabulate , Edit/Jam-mode"
.text "Shift+N , C=+N: Instr.Name, Tbl.NoteMode"
.text "C=+,/. (Sh)+/-: Select SubTune/Ins./Patt"
.text "RETURN in patt: NoteOFF/ON(Sh), Len.(C=)"
.text "RET.in orderl.: Select Patterns to edit "
.text "C=+0..9 Sh+1/2: Select Octave / Mute chn"
.text "C=+F/G, C=+A/J: FrameSpd/AutoAdv/FullJam"
.text "Sh+Q/W, C=+Q/W: Transpose Patt from Curs"
.text "C= + X / C / V: Cut/Copy/Paste from Curs"
.text "Sh/C=+V/U/D/L/T/S:Vib/Sli/Lega/Port/Sync"
.text "---wf-------frq--------vol--fx-0:Duratio"
.text "Xx:X&Wait 0x:Wait   0x:Wait  123:Sli/Por"
.text "2x:Pulse 1x2X:Slide 1x:FadeIn  4:WF.jump"
.text "3x:PSync 3x:Detune  2x:FadeOut 5:NoteVol"
.text "4x:Noise 4x5x:RelUp 3x:PrioVol 6:VOLjump"
.text "5x:NSync 60:IncVibr 4x:LoudVol 7:FRQjump"
.text "6x:NoSnd 7x6x:RelDn 5x:AverVol 8:Vibrato"
.text "80..FE:  80:Vibrato -orderlist-9:Detune-"
.text "   Phase 81..FE:Abs 8,9:Transp,A:MainVol"
.text "FF:Jump(everywh)D:Detun,E:Fade,B:TempoPr"
.bend
helptxt_end
;.enc none
chrmod .var ()
.for i=0,i<helptxt_end-helptxt,i=i+1
 .if (helptxt[i]>=$60)
  chrmod .var chrmod..(helptxt[i]&$1F,)
 .else
  chrmod .var chrmod..(helptxt[i],)
 .fi
.next
*=helptxt
.byte chrmod
.fi ;end of HELPTEXT_SUPPORT

;=========================== EXAMPLE TUNE ==============================
.if (EXAMPLE_TUNE_SUPPORT>0)
demotune
 .if (PLAYERTYPE==0) ;normal?
  .binary "examplnormex.tzw",2
 .elsif (PLAYERTYPE==1) ;medium?
  .binary "examplmedium.tzw",2
 .elsif (PLAYERTYPE==2) ;light?
  .binary "examplelight.tzw",2
 .elsif (PLAYERTYPE==3) ;extra?
  .binary "examplnormex.tzw",2
 .elsif (PLAYERTYPE==4) ;bare?
  .binary "example-bare.tzw",2
 .elsif (PLAYERTYPE==5) ;demo?
  .binary "example-demo.tzw",2
 .fi
demotune_end
.fi
;.if (MACHINE<2)
; .cerror (*>=$FC00),"TUNEDATA OVERLAPPING NON-BANKING ROM",*
;.fi


;========================== MISC DATA UNDER I/O ========================
.if (PLAYABLE_EXPORT_SUPPORT>0 || RUNNABLE_EXPORT_SUPPORT>0)
 .if (MACHINE==2)
  #relocdata
 .fi
.fi


.if (MACHINE==0)
end_of_code
 .cerror (*>=$4000),"Exceeded Top of C16 memory!",* ;top of C16 memory reached? no way to run :(
.elsif (MACHINE==1)
end_of_code
 .cerror (*>=$FD00),"Exceeded Top Of memory!",*
.else
 *=end_of_code
 .cerror (*>=$FFF8),"Exceeded Top Of memory!",*
.fi




;=======================================================================
